@@ -11,12 +11,12 @@
 
 #include "llvm/ADT/DenseMap.h"
 
-namespace lld {
-namespace coff {
+namespace lld::coff {
 class SectionChunk;
+class COFFLinkerContext;
 
-llvm::DenseMap<const SectionChunk *, int> computeCallGraphProfileOrder();
-} // namespace coff
-} // namespace lld
+llvm::DenseMap<const SectionChunk *, int>
+computeCallGraphProfileOrder(COFFLinkerContext &ctx);
+} // namespace lld::coff
 
 #endif
